@@ -40,6 +40,10 @@
 #include <string.h>
 #include <assert.h>
 
+#ifdef HAVE_FNMATCH_H
+#include <fnmatch.h>
+#endif
+
 #ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
 #endif
@@ -62,6 +66,7 @@
 
 #include "ascii-string.h"
 #include "cleanup.h"
+#include "strndup.h"
 
 /* -D ip.rules=1 to enable debugging of rules and rule matching. */
 NBDKIT_DLL_PUBLIC int ip_debug_rules;
@@ -70,13 +75,14 @@ struct rule {
   struct rule *next;
   enum { BAD = 0,
          ANY, ANYV4, ANYV6, IPV4, IPV6,
-         ANYUNIX, PID, UID, GID, SECURITY,
+         ANYUNIX, DN, ISSUER_DN, PID, UID, GID, SECURITY,
          ANYVSOCK, VSOCKCID, VSOCKPORT } type;
   union {
     struct in_addr ipv4;        /* for IPV4, IPV6 */
     struct in6_addr ipv6;
     int64_t id;                 /* for PID, UID, GID, VSOCKCID, VSOCKPORT */
-    const char *label;          /* for SECURITY */
+    char *label;                /* for SECURITY (must be freed) */
+    const char *glob;           /* for DN or ISSUER_DN */
   } u;
   unsigned prefixlen;           /* for IPV4, IPV6 */
 };
@@ -84,6 +90,11 @@ struct rule {
 static struct rule *allow_rules, *allow_rules_last;
 static struct rule *deny_rules, *deny_rules_last;
 
+/* If there are any dn:/issuer-dn: rules then we must do more
+ * expensive "late filtering" in the .open method.
+ */
+static bool late_filtering = false;
+
 static void
 print_rule (const char *name, const struct rule *rule, const char *suffix)
 {
@@ -118,6 +129,12 @@ print_rule (const char *name, const struct rule *rule, const char *suffix)
   case ANYUNIX:
     nbdkit_debug ("%s=anyunix%s", name, suffix);
     break;
+  case DN:
+    nbdkit_debug ("%s=dn:%s%s", name, rule->u.glob, suffix);
+    break;
+  case ISSUER_DN:
+    nbdkit_debug ("%s=issuer-dn:%s%s", name, rule->u.glob, suffix);
+    break;
   case PID:
     nbdkit_debug ("%s=pid:%" PRIi64 "%s", name, rule->u.id, suffix);
     break;
@@ -165,6 +182,8 @@ free_rules (struct rule *rules)
 
   for (rule = rules; rule != NULL; rule = next) {
     next = rule->next;
+    if (rule->type == SECURITY)
+      free (rule->u.label);
     free (rule);
   }
 }
@@ -276,6 +295,18 @@ parse_rule (const char *paramname,
     return 0;
   }
 
+  if (n >= 3 && ascii_strncasecmp (value, "dn:", 3) == 0) {
+    late_filtering = true;
+    new_rule->type = DN;
+    new_rule->u.glob = &value[3];
+    return 0;
+  }
+  if (n >= 10 && ascii_strncasecmp (value, "issuer-dn:", 10) == 0) {
+    late_filtering = true;
+    new_rule->type = ISSUER_DN;
+    new_rule->u.glob = &value[10];
+    return 0;
+  }
   if (n >= 4 && ascii_strncasecmp (value, "pid:", 4) == 0) {
     new_rule->type = PID;
     if (nbdkit_parse_int64_t ("pid:", &value[4], &new_rule->u.id) == -1)
@@ -309,7 +340,11 @@ parse_rule (const char *paramname,
 
   if (n >= 9 && ascii_strncasecmp (value, "security:", 9) == 0) {
     new_rule->type = SECURITY;
-    new_rule->u.label = &value[9];
+    new_rule->u.label = strndup (&value[9], n-9);
+    if (new_rule->u.label == NULL) {
+      nbdkit_error ("strndup: %m");
+      return -1;
+    }
     return 0;
   }
 
@@ -400,6 +435,11 @@ parse_rules (const char *paramname,
 {
   size_t n;
 
+  /* Special exception for dn:/issuer-dn: since they contain commas. */
+  if (ascii_strncasecmp (value, "dn:", 3) == 0 ||
+      ascii_strncasecmp (value, "issuer-dn:", 10) == 0)
+    return parse_rule (paramname, rules, rules_last, value, strlen (value));
+
   while (*value != '\0') {
     n = strcspn (value, ",");
     if (n == 0) {
@@ -482,15 +522,53 @@ ipv6_equal (struct in6_addr addr1, struct in6_addr addr2, unsigned prefixlen)
   return true;
 }
 
+#ifdef HAVE_FNMATCH_H
+static bool
+dn_matches (const char *name, const char *dn, const char *glob)
+{
+  int r;
+  bool ret;
+
+  if (dn == NULL)
+    return false;
+
+  if (ip_debug_rules)
+    nbdkit_debug ("ip: %s = \"%s\"", name, dn);
+
+  r = fnmatch (glob, dn, FNM_CASEFOLD);
+  switch (r) {
+  case 0:
+    ret = true;
+    break;
+  case FNM_NOMATCH:
+    ret = false;
+    break;
+  default:
+    /* I even looked at the glibc code to see what the unspecified
+     * "errors" were and it was not clear.  The best we can do here is
+     * report the code and errno, but continue as if it doesn't match.
+     */
+    nbdkit_error ("fnmatch returned error code %d: %m", r);
+    ret = false;
+  }
+
+  return ret;
+}
+#endif
+
 static bool
 matches_rule (const struct rule *rule,
-              int family, const struct sockaddr *addr)
+              const struct sockaddr *addr)
 {
+  const int family = ((struct sockaddr_in *)addr)->sin_family;
   const struct sockaddr_in *sin;
   uint32_t cin, rin, mask;
   const struct sockaddr_in6 *sin6;
 #if defined (AF_VSOCK) && defined (VMADDR_CID_ANY)
   const struct sockaddr_vm *svm;
+#endif
+#ifdef HAVE_FNMATCH_H
+  char *dn;
 #endif
   char *label;
   bool b;
@@ -521,6 +599,26 @@ matches_rule (const struct rule *rule,
   case ANYUNIX:
     return family == AF_UNIX;
 
+  case DN:
+#ifdef HAVE_FNMATCH_H
+    dn = nbdkit_peer_tls_dn ();
+    b = dn_matches ("TLS DN", dn, rule->u.glob);
+    free (dn);
+    return b;
+#else
+    return false;
+#endif
+
+  case ISSUER_DN:
+#ifdef HAVE_FNMATCH_H
+    dn = nbdkit_peer_tls_issuer_dn ();
+    b = dn_matches ("TLS issuer DN", dn, rule->u.glob);
+    free (dn);
+    return b;
+#else
+    return false;
+#endif
+
     /* Note these work even if the underlying nbdkit_peer_* call fails. */
   case PID:
     if (family != AF_UNIX) return false;
@@ -580,13 +678,13 @@ matches_rule (const struct rule *rule,
 
 static bool
 matches_rules_list (const char *name, const struct rule *rules,
-                    int family, const struct sockaddr *addr)
+                    const struct sockaddr *addr)
 {
   const struct rule *rule;
   bool b;
 
   for (rule = rules; rule != NULL; rule = rule->next) {
-    b = matches_rule (rule, family, addr);
+    b = matches_rule (rule, addr);
     if (ip_debug_rules)
       print_rule (name, rule, b ? " => yes" : " => no");
     if (b)
@@ -596,32 +694,6 @@ matches_rules_list (const char *name, const struct rule *rules,
   return false;
 }
 
-static bool
-check_if_allowed (const struct sockaddr *addr)
-{
-  int family = ((struct sockaddr_in *)addr)->sin_family;
-
-  /* There's an implicit allow all for non-IP, non-Unix, non-AF_VSOCK
-   * sockets, see the manual.
-   */
-  if (family != AF_INET && family != AF_INET6 && family != AF_UNIX
-#if defined (AF_VSOCK) && defined (VMADDR_CID_ANY)
-      && family != AF_VSOCK
-#endif
-      )
-    return true;
-
-  if (matches_rules_list ("ip: match client with allow",
-                          allow_rules, family, addr))
-    return true;
-
-  if (matches_rules_list ("ip: match client with deny",
-                          deny_rules, family, addr))
-    return false;
-
-  return true;
-}
-
 /* Print the peer name in preconnect, when -D ip.rules=1 */
 static void
 print_peer_name (const struct sockaddr *sa)
@@ -672,23 +744,40 @@ print_peer_name (const struct sockaddr *sa)
   }
 }
 
-static int
-ip_preconnect (nbdkit_next_preconnect *next, nbdkit_backend *nxdata,
-               int readonly)
+static bool
+check_if_allowed (void)
 {
   struct sockaddr_storage addr;
   socklen_t addrlen = sizeof addr;
 
   if (nbdkit_peer_name ((struct sockaddr *) &addr, &addrlen) == -1)
-    return -1;                  /* We should fail closed ... */
+    return false;               /* We should fail closed ... */
   if (ip_debug_rules)
     print_peer_name ((struct sockaddr *) &addr);
 
-  /* Follow the rules. */
-  if (check_if_allowed ((struct sockaddr *) &addr) == false) {
-    nbdkit_error ("client not permitted to connect "
-                  "because of source address restriction");
-    return -1;
+  if (matches_rules_list ("ip: match client with allow",
+                          allow_rules, (struct sockaddr *) &addr))
+    return true;
+
+  if (matches_rules_list ("ip: match client with deny",
+                          deny_rules, (struct sockaddr *) &addr))
+    return false;
+
+  return true;
+}
+
+/* Normal/early filtering, if there are no dn:/issuer-dn: rules. */
+static int
+ip_preconnect (nbdkit_next_preconnect *next, nbdkit_backend *nxdata,
+               int readonly)
+{
+  if (!late_filtering) {
+    /* Follow the rules. */
+    if (check_if_allowed () == false) {
+      nbdkit_error ("client not permitted to connect "
+                    "because of source address restriction");
+      return -1;
+    }
   }
 
   if (next (nxdata, readonly) == -1)
@@ -697,6 +786,26 @@ ip_preconnect (nbdkit_next_preconnect *next, nbdkit_backend *nxdata,
   return 0;
 }
 
+/* Late filtering, if there are dn:/issuer-dn: rules. */
+static void *
+ip_open (nbdkit_next_open *next, nbdkit_context *nxdata,
+         int readonly, const char *exportname, int is_tls)
+{
+  if (late_filtering) {
+    /* Follow the rules. */
+    if (check_if_allowed () == false) {
+      nbdkit_error ("client not permitted to connect "
+                    "because of source address restriction");
+      return NULL;
+    }
+  }
+
+  if (next (nxdata, readonly, exportname) == -1)
+    return NULL;
+
+  return NBDKIT_HANDLE_NOT_NEEDED;
+}
+
 static int
 ip_thread_model (void)
 {
@@ -712,6 +821,7 @@ static struct nbdkit_filter filter = {
   .config_complete   = ip_config_complete,
   .config_help       = ip_config_help,
   .preconnect        = ip_preconnect,
+  .open              = ip_open,
 };
 
 NBDKIT_REGISTER_FILTER (filter)
