@@ -39,12 +39,14 @@
 #include <errno.h>
 
 #include <caml/alloc.h>
+#include <caml/bigarray.h>
 #include <caml/callback.h>
 #include <caml/fail.h>
 #include <caml/memory.h>
 #include <caml/mlvalues.h>
 #include <caml/printexc.h>
 #include <caml/threads.h>
+#include <caml/version.h>
 
 #define NBDKIT_API_VERSION 2
 #include <nbdkit-plugin.h>
@@ -138,6 +140,19 @@ plugin_init (void)
  * next time.
  */
 
+/* This macro calls nbdkit_error when we get an exception thrown in
+ * OCaml callback code.  The 'return_stmt' parameter is usually a call
+ * to CAMLreturnT, but may be empty in order to fall through.
+ */
+#define EXCEPTION_TO_ERROR(rv, return_stmt)                             \
+  do {                                                                  \
+    if (Is_exception_result (rv)) {                                     \
+      nbdkit_error ("%s: %s", __func__,                                 \
+                    caml_format_exception (Extract_exception (rv)));    \
+      return_stmt;                                                      \
+    }                                                                   \
+  } while (0)
+
 static void
 load_wrapper (void)
 {
@@ -175,9 +190,12 @@ dump_plugin_wrapper (void)
   CAMLparam0 ();
   CAMLlocal1 (rv);
 
+  printf ("ocaml_version_major=%d\n", OCAML_VERSION_MAJOR);
+  printf ("ocaml_version_minor=%d\n", OCAML_VERSION_MINOR);
+  printf ("ocaml_version=%s\n", OCAML_VERSION_STRING);
+
   rv = caml_callback_exn (dump_plugin_fn, Val_unit);
-  if (Is_exception_result (rv))
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+  EXCEPTION_TO_ERROR (rv, /* fallthrough */);
   CAMLreturn0;
 }
 
@@ -193,10 +211,7 @@ config_wrapper (const char *key, const char *val)
   valv = caml_copy_string (val);
 
   rv = caml_callback2_exn (config_fn, keyv, valv);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
@@ -210,10 +225,7 @@ config_complete_wrapper (void)
   CAMLlocal1 (rv);
 
   rv = caml_callback_exn (config_complete_fn, Val_unit);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
@@ -227,10 +239,7 @@ thread_model_wrapper (void)
   CAMLlocal1 (rv);
 
   rv = caml_callback_exn (thread_model_fn, Val_unit);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Int_val (rv));
 }
@@ -244,10 +253,7 @@ get_ready_wrapper (void)
   CAMLlocal1 (rv);
 
   rv = caml_callback_exn (get_ready_fn, Val_unit);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
@@ -261,10 +267,7 @@ after_fork_wrapper (void)
   CAMLlocal1 (rv);
 
   rv = caml_callback_exn (after_fork_fn, Val_unit);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
@@ -278,10 +281,7 @@ cleanup_wrapper (void)
   CAMLlocal1 (rv);
 
   rv = caml_callback_exn (cleanup_fn, Val_unit);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturn0;
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturn0);
 
   CAMLreturn0;
 }
@@ -295,10 +295,7 @@ preconnect_wrapper (int readonly)
   CAMLlocal1 (rv);
 
   rv = caml_callback_exn (preconnect_fn, Val_bool (readonly));
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
@@ -313,10 +310,7 @@ list_exports_wrapper (int readonly, int is_tls, struct nbdkit_exports *exports)
 
   rv = caml_callback2_exn (list_exports_fn, Val_bool (readonly),
                            Val_bool (is_tls));
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   /* Convert exports list into calls to nbdkit_add_export. */
   while (rv != Val_emptylist) {
@@ -347,15 +341,17 @@ default_export_wrapper (int readonly, int is_tls)
 
   rv = caml_callback2_exn (default_export_fn, Val_bool (readonly),
                            Val_bool (is_tls));
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (const char *, NULL);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (const char *, NULL));
 
   name = nbdkit_strdup_intern (String_val (rv));
   CAMLreturnT (const char *, name);
 }
 
+/* The C handle. */
+struct handle {
+  value v;       /* The OCaml handle, also registered as a GC root. */
+};
+
 static void *
 open_wrapper (int readonly)
 {
@@ -363,43 +359,41 @@ open_wrapper (int readonly)
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
-  value *ret;
+  struct handle *h;
 
   rv = caml_callback_exn (open_fn, Val_bool (readonly));
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (void *, NULL);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (void *, NULL));
 
   /* Allocate a root on the C heap that points to the OCaml handle. */
-  ret = malloc (sizeof *ret);
-  if (ret == NULL) abort ();
-  *ret = rv;
-  caml_register_generational_global_root (ret);
+  h = malloc (sizeof *h);
+  if (h == NULL) {
+    nbdkit_error ("malloc: %m");
+    CAMLreturnT (void *, NULL);
+  }
+  h->v = rv;
+  caml_register_generational_global_root (&h->v);
 
-  CAMLreturnT (void *, ret);
+  CAMLreturnT (void *, h);
 }
 
 /* We always have a close wrapper, since we need to unregister the
  * global root, free the handle and unregister the thread.
  */
 static void
-close_wrapper (void *h)
+close_wrapper (void *hv)
 {
   caml_c_thread_register ();
   do_caml_acquire_runtime_system ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
   if (close_fn) {
-    rv = caml_callback_exn (close_fn, *(value *) h);
-    if (Is_exception_result (rv)) {
-      nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-      /*FALLTHROUGH*/
-    }
+    rv = caml_callback_exn (close_fn, h->v);
+    EXCEPTION_TO_ERROR (rv, /* fallthrough */);
   }
 
-  caml_remove_generational_global_root (h);
+  caml_remove_generational_global_root (&h->v);
   free (h);
   do_caml_release_runtime_system ();
   caml_c_thread_unregister ();
@@ -408,59 +402,53 @@ close_wrapper (void *h)
 }
 
 static const char *
-export_description_wrapper (void *h)
+export_description_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
   const char *desc;
 
-  rv = caml_callback_exn (export_description_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (const char *, NULL);
-  }
+  rv = caml_callback_exn (export_description_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (const char *, NULL));
 
   desc = nbdkit_strdup_intern (String_val (rv));
   CAMLreturnT (const char *, desc);
 }
 
 static int64_t
-get_size_wrapper (void *h)
+get_size_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
   int64_t r;
 
-  rv = caml_callback_exn (get_size_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int64_t, -1);
-  }
+  rv = caml_callback_exn (get_size_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int64_t, -1));
 
   r = Int64_val (rv);
   CAMLreturnT (int64_t, r);
 }
 
 static int
-block_size_wrapper (void *h,
+block_size_wrapper (void *hv,
                     uint32_t *minimum, uint32_t *preferred, uint32_t *maximum)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
   int i;
   int64_t i64;
 
-  rv = caml_callback_exn (block_size_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (block_size_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   i = Int_val (Field  (rv, 0));
   if (i < 0 || i > 65536) {
@@ -490,171 +478,151 @@ block_size_wrapper (void *h,
 }
 
 static int
-can_write_wrapper (void *h)
+can_write_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (can_write_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (can_write_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Bool_val (rv));
 }
 
 static int
-can_flush_wrapper (void *h)
+can_flush_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (can_flush_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (can_flush_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Bool_val (rv));
 }
 
 static int
-is_rotational_wrapper (void *h)
+is_rotational_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (is_rotational_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (is_rotational_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Bool_val (rv));
 }
 
 static int
-can_trim_wrapper (void *h)
+can_trim_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (can_trim_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (can_trim_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Bool_val (rv));
 }
 
 static int
-can_zero_wrapper (void *h)
+can_zero_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (can_zero_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (can_zero_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Bool_val (rv));
 }
 
 static int
-can_fua_wrapper (void *h)
+can_fua_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (can_fua_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (can_fua_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Int_val (rv));
 }
 
 static int
-can_fast_zero_wrapper (void *h)
+can_fast_zero_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (can_fast_zero_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (can_fast_zero_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Bool_val (rv));
 }
 
 static int
-can_cache_wrapper (void *h)
+can_cache_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (can_cache_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (can_cache_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Int_val (rv));
 }
 
 static int
-can_extents_wrapper (void *h)
+can_extents_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (can_extents_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (can_extents_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Bool_val (rv));
 }
 
 static int
-can_multi_conn_wrapper (void *h)
+can_multi_conn_wrapper (void *hv)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal1 (rv);
+  struct handle *h = hv;
 
-  rv = caml_callback_exn (can_multi_conn_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback_exn (can_multi_conn_fn, h->v);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, Bool_val (rv));
 }
@@ -688,142 +656,132 @@ Val_flags (uint32_t flags)
   CAMLreturn (rv);
 }
 
+/* Wrap the buf in an OCaml bigarray so that callers can read and
+ * write directly to it without copying.
+ * https://discuss.ocaml.org/t/is-there-a-simple-library-that-wraps-a-c-mallocd-char-buffer/14323
+ * https://v2.ocaml.org/manual/intfc.html#ss:C-Bigarrays-wrap
+ */
 static int
-pread_wrapper (void *h, void *buf, uint32_t count, uint64_t offset,
+pread_wrapper (void *hv, void *buf, uint32_t count, uint64_t offset,
                uint32_t flags)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
-  CAMLlocal4 (rv, countv, offsetv, flagsv);
-  mlsize_t len;
+  CAMLlocal4 (rv, ba, offsetv, flagsv);
+  struct handle *h = hv;
+  long dims[1] = { count };
 
-  countv = Val_int (count);
+  ba = caml_ba_alloc (CAML_BA_CHAR|CAML_BA_C_LAYOUT, 1, buf, dims);
   offsetv = caml_copy_int64 (offset);
   flagsv = Val_flags (flags);
 
-  value args[] = { *(value *) h, countv, offsetv, flagsv };
+  value args[] = { h->v, ba, offsetv, flagsv };
   rv = caml_callbackN_exn (pread_fn, ARRAY_SIZE (args), args);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
-  len = caml_string_length (rv);
-  if (len < count) {
-    nbdkit_error ("buffer returned from pread is too small");
-    CAMLreturnT (int, -1);
-  }
-
-  memcpy (buf, String_val (rv), count);
   CAMLreturnT (int, 0);
 }
 
 static int
-pwrite_wrapper (void *h, const void *buf, uint32_t count, uint64_t offset,
+pwrite_wrapper (void *hv, const void *buf, uint32_t count, uint64_t offset,
                 uint32_t flags)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
-  CAMLlocal4 (rv, strv, offsetv, flagsv);
+  CAMLlocal4 (rv, ba, offsetv, flagsv);
+  struct handle *h = hv;
+  long dims[1] = { count };
 
-  strv = caml_alloc_initialized_string (count, buf);
+  /* We discard the const of the incoming buffer, and in theory OCaml
+   * plugins could try writing to it. XXX
+   */
+  ba = caml_ba_alloc (CAML_BA_CHAR|CAML_BA_C_LAYOUT, 1, (void *) buf, dims);
   offsetv = caml_copy_int64 (offset);
   flagsv = Val_flags (flags);
 
-  value args[] = { *(value *) h, strv, offsetv, flagsv };
+  value args[] = { h->v, ba, offsetv, flagsv };
   rv = caml_callbackN_exn (pwrite_fn, ARRAY_SIZE (args), args);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
 
 static int
-flush_wrapper (void *h, uint32_t flags)
+flush_wrapper (void *hv, uint32_t flags)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal2 (rv, flagsv);
+  struct handle *h = hv;
 
   flagsv = Val_flags (flags);
 
-  rv = caml_callback2_exn (flush_fn, *(value *) h, flagsv);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  rv = caml_callback2_exn (flush_fn, h->v, flagsv);
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
 
 static int
-trim_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags)
+trim_wrapper (void *hv, uint32_t count, uint64_t offset, uint32_t flags)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal4 (rv, countv, offsetv, flagsv);
+  struct handle *h = hv;
 
   countv = caml_copy_int64 (count);
   offsetv = caml_copy_int64 (offset);
   flagsv = Val_flags (flags);
 
-  value args[] = { *(value *) h, countv, offsetv, flagsv };
+  value args[] = { h->v, countv, offsetv, flagsv };
   rv = caml_callbackN_exn (trim_fn, ARRAY_SIZE (args), args);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
 
 static int
-zero_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags)
+zero_wrapper (void *hv, uint32_t count, uint64_t offset, uint32_t flags)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal4 (rv, countv, offsetv, flagsv);
+  struct handle *h = hv;
 
   countv = caml_copy_int64 (count);
   offsetv = caml_copy_int64 (offset);
   flagsv = Val_flags (flags);
 
-  value args[] = { *(value *) h, countv, offsetv, flagsv };
+  value args[] = { h->v, countv, offsetv, flagsv };
   rv = caml_callbackN_exn (zero_fn, ARRAY_SIZE (args), args);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
 
 static int
-extents_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags,
+extents_wrapper (void *hv, uint32_t count, uint64_t offset, uint32_t flags,
                  struct nbdkit_extents *extents)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal5 (rv, countv, offsetv, flagsv, v);
+  struct handle *h = hv;
 
   countv = caml_copy_int64 (count);
   offsetv = caml_copy_int64 (offset);
   flagsv = Val_flags (flags);
 
-  value args[] = { *(value *) h, countv, offsetv, flagsv };
+  value args[] = { h->v, countv, offsetv, flagsv };
   rv = caml_callbackN_exn (extents_fn, ARRAY_SIZE (args), args);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   /* Convert extents list into calls to nbdkit_add_extent. */
   while (rv != Val_emptylist) {
@@ -848,23 +806,21 @@ extents_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags,
 }
 
 static int
-cache_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags)
+cache_wrapper (void *hv, uint32_t count, uint64_t offset, uint32_t flags)
 {
   caml_c_thread_register ();
   ACQUIRE_RUNTIME_FOR_CURRENT_SCOPE ();
   CAMLparam0 ();
   CAMLlocal4 (rv, countv, offsetv, flagsv);
+  struct handle *h = hv;
 
   countv = caml_copy_int64 (count);
   offsetv = caml_copy_int64 (offset);
   flagsv = Val_flags (flags);
 
-  value args[] = { *(value *) h, countv, offsetv, flagsv };
+  value args[] = { h->v, countv, offsetv, flagsv };
   rv = caml_callbackN_exn (cache_fn, ARRAY_SIZE (args), args);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    CAMLreturnT (int, -1);
-  }
+  EXCEPTION_TO_ERROR (rv, CAMLreturnT (int, -1));
 
   CAMLreturnT (int, 0);
 }
